@@ -21,6 +21,7 @@ use bit_set::BitSet;
 use schemars::JsonSchema;
 use serde::{Deserialize, Serialize};
 use std::borrow::Cow;
+use std::collections::HashSet;
 use thiserror::Error;
 
 /// A rule object to find matching AST nodes. We have three categories of rules in ast-grep.
@@ -179,6 +180,7 @@ impl<L: Language> Rule<L> {
     matches!(self, All(_) | Any(_) | Not(_) | Matches(_))
   }
 
+  /// Check if it has a cyclic referent rule with the id.
   pub(crate) fn check_cyclic(&self, id: &str) -> bool {
     match self {
       Rule::All(all) => all.inner().iter().any(|r| r.check_cyclic(id)),
@@ -191,6 +193,22 @@ impl<L: Language> Rule<L> {
       }
     }
   }
+
+  pub fn defined_vars(&self) -> HashSet<&str> {
+    match self {
+      Rule::Pattern(p) => p.defined_vars(),
+      Rule::Kind(_) => HashSet::new(),
+      Rule::Regex(_) => HashSet::new(),
+      Rule::Has(c) => c.defined_vars(),
+      Rule::Inside(p) => p.defined_vars(),
+      Rule::Precedes(f) => f.defined_vars(),
+      Rule::Follows(f) => f.defined_vars(),
+      Rule::All(sub) => sub.inner().iter().flat_map(|r| r.defined_vars()).collect(),
+      Rule::Any(sub) => sub.inner().iter().flat_map(|r| r.defined_vars()).collect(),
+      Rule::Not(sub) => sub.inner().defined_vars(),
+      Rule::Matches(r) => r.defined_vars(),
+    }
+  }
 }
 
 impl<L: Language> Matcher<L> for Rule<L> {
@@ -271,6 +289,8 @@ pub enum RuleSerializeError {
   MatchesReference(#[from] ReferentRuleError),
   #[error("field is only supported in has/inside.")]
   FieldNotSupported,
+  #[error("Relational rule contains invalid field {0}.")]
+  InvalidField(String),
 }
 
 // TODO: implement positive/non positive
@@ -503,4 +523,17 @@ inside:
     assert!(rules[0].is_atomic());
     assert!(rules[1].is_relational());
   }
+
+  #[test]
+  fn test_defined_vars() {
+    let src = r"
+pattern: var $A = 123
+inside:
+  pattern: var $B = 456
+";
+    let rule: SerializableRule = from_str(src).expect("cannot parse rule");
+    let env = DeserializeEnv::new(TypeScript::Tsx);
+    let rule = deserialize_rule(rule, &env).expect("should deserialize");
+    assert_eq!(rule.defined_vars(), ["A", "B"].into_iter().collect());
+  }
 }
