@@ -91,14 +91,45 @@ impl<L: Language> SerializableRuleConfig<L> {
         .with_globals(globals)
         .with_rewriters(&rewriters);
       let env = self.get_deserialize_env(env)?;
-      let env = val.core.get_deserialize_env(env)?;
-      let rewriter = val.core.get_matcher_from_env(&env)?;
+      let rewriter = val.core.get_matcher(env)?;
       rewriters.insert(&val.id, rewriter).expect("should work");
     }
+    check_rewriters_in_transform(&rule, &rewriters)?;
     Ok(rule)
   }
 }
 
+fn check_rewriters_in_transform<L: Language>(
+  rule: &RuleCore<L>,
+  rewriters: &GlobalRules<L>,
+) -> Result<(), RuleConfigError> {
+  let rewriters = rewriters.read();
+  if let Some(err) = check_one_rewriter_in_rule(rule, &rewriters) {
+    return Err(err);
+  }
+  let error = rewriters
+    .values()
+    .find_map(|rewriter| check_one_rewriter_in_rule(rewriter, &rewriters));
+  if let Some(err) = error {
+    return Err(err);
+  }
+  Ok(())
+}
+
+fn check_one_rewriter_in_rule<L: Language>(
+  rule: &RuleCore<L>,
+  rewriters: &HashMap<String, RuleCore<L>>,
+) -> Option<RuleConfigError> {
+  let transform = rule.transform.as_ref()?;
+  let mut used_rewriters = transform
+    .values()
+    .flat_map(|trans| trans.used_rewriters().iter());
+  let undefined_writers = used_rewriters.find(|r| !rewriters.contains_key(*r))?;
+  Some(RuleConfigError::UndefinedRewriter(
+    undefined_writers.to_string(),
+  ))
+}
+
 impl<L: Language> Deref for SerializableRuleConfig<L> {
   type Target = SerializableRuleCore;
   fn deref(&self) -> &Self::Target {
@@ -407,7 +438,7 @@ transform:
   B:
     rewrite:
       rewriters: [re]
-      source: $A
+      source: $B
 rewriters:
 - id: re
   rule: {matches: num}
@@ -478,4 +509,51 @@ rewriters:
     let b = nm.get_env().get_transformed("B").expect("should have");
     assert_eq!(String::from_utf8_lossy(b), "1919810");
   }
+
+  fn make_undefined_error(src: &str) -> String {
+    let rule: SerializableRuleConfig<TypeScript> = from_str(src).expect("should parse");
+    let err = RuleConfig::try_from(rule, &Default::default());
+    match err {
+      Err(RuleConfigError::UndefinedRewriter(name)) => name,
+      _ => panic!("unexpected parsing result"),
+    }
+  }
+
+  #[test]
+  fn test_undefined_rewriter() {
+    let undefined = make_undefined_error(
+      r"
+id: test
+rule: {pattern: 'a = $A'}
+language: Tsx
+transform:
+  B: { rewrite: { rewriters: [not-defined], source: $A } }
+rewriters:
+- id: re
+  rule: {kind: number, pattern: $A}
+  fix: hah
+    ",
+    );
+    assert_eq!(undefined, "not-defined");
+  }
+
+  #[test]
+  fn test_undefined_rewriter_in_transform() {
+    let undefined = make_undefined_error(
+      r"
+id: test
+rule: {pattern: 'a = $A'}
+language: Tsx
+transform:
+  B: { rewrite: { rewriters: [re], source: $A } }
+rewriters:
+- id: re
+  rule: {kind: number, pattern: $A}
+  transform:
+    C: { rewrite: { rewriters: [nested-undefined], source: $A } }
+  fix: hah
+    ",
+    );
+    assert_eq!(undefined, "nested-undefined");
+  }
 }
