@@ -16,7 +16,7 @@ use schemars::JsonSchema;
 use thiserror::Error;
 
 use std::borrow::Cow;
-use std::collections::HashMap;
+use std::collections::{HashMap, HashSet};
 use std::ops::Deref;
 
 #[derive(Debug, Error)]
@@ -31,6 +31,10 @@ pub enum RuleConfigError {
   Fixer(#[from] FixerError),
   #[error("constraints is not configured correctly.")]
   Constraints(#[source] RuleSerializeError),
+  #[error("Undefined meta var {0} used in {1}.")]
+  UndefinedMetaVar(String, &'static str),
+  #[error("Undefined rewriter {0} used in transform.")]
+  UndefinedRewriter(String),
 }
 
 type RResult<T> = std::result::Result<T, RuleConfigError>;
@@ -69,7 +73,7 @@ impl SerializableRuleCore {
     }
   }
 
-  fn get_meta_var_matchers<L: Language>(
+  fn get_constraints<L: Language>(
     &self,
     env: &DeserializeEnv<L>,
   ) -> RResult<HashMap<String, Rule<L>>> {
@@ -92,15 +96,14 @@ impl SerializableRuleCore {
     }
   }
 
-  // TODO: this is wrong, it does not register local utils to the env
-  pub(crate) fn get_matcher_from_env<L: Language>(
-    &self,
-    env: &DeserializeEnv<L>,
-  ) -> RResult<RuleCore<L>> {
+  fn get_matcher_from_env<L: Language>(&self, env: &DeserializeEnv<L>) -> RResult<RuleCore<L>> {
     let rule = env.deserialize_rule(self.rule.clone())?;
-    let matchers = self.get_meta_var_matchers(env)?;
+    let matchers = self.get_constraints(env)?;
+    let vars = Self::check_var_in_constraints(&rule, &matchers)?;
+    let vars = Self::check_var_in_transform(vars, &self.transform)?;
     let transform = self.transform.clone();
     let fixer = self.get_fixer(env)?;
+    Self::check_var_in_fix(vars, &fixer)?;
     Ok(
       RuleCore::new(rule)
         .with_matchers(matchers)
@@ -110,6 +113,62 @@ impl SerializableRuleCore {
     )
   }
 
+  fn check_var_in_constraints<'r, L: Language>(
+    rule: &'r Rule<L>,
+    constraints: &'r HashMap<String, Rule<L>>,
+  ) -> RResult<HashSet<&'r str>> {
+    let mut vars = rule.defined_vars();
+    for rule in constraints.values() {
+      for var in rule.defined_vars() {
+        vars.insert(var);
+      }
+    }
+    for var in constraints.keys() {
+      let var: &str = var;
+      if !vars.contains(var) {
+        return Err(RuleConfigError::UndefinedMetaVar(
+          var.to_owned(),
+          "constraints",
+        ));
+      }
+    }
+    Ok(vars)
+  }
+
+  fn check_var_in_transform<'r>(
+    mut vars: HashSet<&'r str>,
+    transform: &'r Option<HashMap<String, Transformation>>,
+  ) -> RResult<HashSet<&'r str>> {
+    let Some(transform) = transform else {
+      return Ok(vars);
+    };
+    for var in transform.keys() {
+      vars.insert(var);
+    }
+    for trans in transform.values() {
+      let needed = trans.used_vars();
+      if !vars.contains(needed) {
+        return Err(RuleConfigError::UndefinedMetaVar(
+          needed.to_string(),
+          "transform",
+        ));
+      }
+    }
+    Ok(vars)
+  }
+
+  fn check_var_in_fix<L: Language>(vars: HashSet<&str>, fixer: &Option<Fixer<L>>) -> RResult<()> {
+    let Some(fixer) = fixer else {
+      return Ok(());
+    };
+    for var in fixer.used_vars() {
+      if !vars.contains(&var) {
+        return Err(RuleConfigError::UndefinedMetaVar(var.to_string(), "fix"));
+      }
+    }
+    Ok(())
+  }
+
   pub fn get_matcher<L: Language>(&self, env: DeserializeEnv<L>) -> RResult<RuleCore<L>> {
     let env = self.get_deserialize_env(env)?;
     self.get_matcher_from_env(&env)
@@ -120,7 +179,7 @@ pub struct RuleCore<L: Language> {
   rule: Rule<L>,
   matchers: HashMap<String, Rule<L>>,
   kinds: Option<BitSet>,
-  transform: Option<HashMap<String, Transformation>>,
+  pub(crate) transform: Option<HashMap<String, Transformation>>,
   pub fixer: Option<Fixer<L>>,
   // this is required to hold util rule reference
   utils: RuleRegistration<L>,
@@ -163,6 +222,21 @@ impl<L: Language> RuleCore<L> {
       registration: self.utils.clone(),
     }
   }
+
+  pub fn defined_vars(&self) -> HashSet<&str> {
+    let mut ret = self.rule.defined_vars();
+    for constraint in self.matchers.values() {
+      for var in constraint.defined_vars() {
+        ret.insert(var);
+      }
+    }
+    if let Some(trans) = &self.transform {
+      for key in trans.keys() {
+        ret.insert(key);
+      }
+    }
+    ret
+  }
 }
 impl<L: Language> Deref for RuleCore<L> {
   type Target = Rule<L>;
@@ -331,4 +405,78 @@ transform:
     let transformed = env.get_transformed("C").expect("should transform");
     assert_eq!(String::from_utf8_lossy(transformed), "a");
   }
+
+  #[test]
+  fn test_defined_vars() {
+    let env = DeserializeEnv::new(TypeScript::Tsx);
+    let ser_rule: SerializableRuleCore = from_str(
+      r"
+rule: {pattern: $A = $B}
+constraints:
+  A: { pattern: $C = $D }
+transform:
+  E:
+    substring:
+      source: $B
+      startCar: 1",
+    )
+    .expect("should deser");
+    let matcher = ser_rule.get_matcher(env).expect("should parse");
+    assert_eq!(
+      matcher.defined_vars(),
+      ["A", "B", "C", "D", "E"].into_iter().collect()
+    );
+  }
+
+  fn get_undefined(src: &str) -> (String, &str) {
+    let env = DeserializeEnv::new(TypeScript::Tsx);
+    let ser_rule: SerializableRuleCore = from_str(src).expect("should deser");
+    match ser_rule.get_matcher(env) {
+      Err(RuleConfigError::UndefinedMetaVar(name, section)) => (name, section),
+      _ => panic!("should error"),
+    }
+  }
+
+  #[test]
+  fn test_undefined_vars_in_constraints() {
+    let (name, section) = get_undefined(
+      r"
+rule: {pattern: $A}
+constraints: {B: {pattern: bbb}}
+",
+    );
+    assert_eq!(name, "B");
+    assert_eq!(section, "constraints");
+  }
+  #[test]
+  fn test_undefined_vars_in_transform() {
+    let (name, section) = get_undefined(
+      r"
+rule: {pattern: $A}
+constraints: {A: {pattern: $C}}
+transform:
+  B:
+    replace: {source: $C, replace: a, by: b }
+  D:
+    replace: {source: $E, replace: a, by: b }
+",
+    );
+    assert_eq!(name, "E");
+    assert_eq!(section, "transform");
+  }
+  #[test]
+  fn test_undefined_vars_in_fix() {
+    let (name, section) = get_undefined(
+      r"
+rule: {pattern: $A}
+constraints: {A: {pattern: $C}}
+transform:
+  B:
+    replace: {source: $C, replace: a, by: b }
+fix: $D
+",
+    );
+    assert_eq!(name, "D");
+    assert_eq!(section, "fix");
+  }
 }
