@@ -142,6 +142,30 @@ impl Transformation {
       T::Rewrite(r) => r.compute(ctx),
     }
   }
+
+  pub fn used_vars(&self) -> &str {
+    fn strip(s: &str) -> &str {
+      s.strip_prefix("$$$").unwrap_or_else(|| &s[1..])
+    }
+    use Transformation as T;
+    // NOTE: meta_var in transform always starts with `$`, for now
+    match self {
+      T::Replace(r) => strip(&r.source),
+      T::Substring(s) => strip(&s.source),
+      T::Convert(c) => strip(&c.source),
+      T::Rewrite(r) => strip(&r.source),
+    }
+  }
+
+  pub fn used_rewriters(&self) -> &[String] {
+    use Transformation as T;
+    match self {
+      T::Replace(_) => &[],
+      T::Substring(_) => &[],
+      T::Convert(_) => &[],
+      T::Rewrite(r) => &r.rewriters,
+    }
+  }
 }
 
 // two lifetime to represent env root lifetime and lang/trans lifetime
