@@ -10,8 +10,8 @@ use serde::{Deserialize, Serialize};
 #[derive(Serialize, Deserialize, Clone, JsonSchema)]
 #[serde(rename_all = "camelCase")]
 pub struct Rewrite {
-  source: String,
-  rewriters: Vec<String>,
+  pub(super) source: String,
+  pub(super) rewriters: Vec<String>,
   // do we need this?
   // sort_by: Option<String>,
   join_by: Option<String>,
