@@ -6,10 +6,12 @@ use crate::source::TSParseError;
 use crate::{Doc, Node, Root, StrDoc};
 
 use bit_set::BitSet;
-use std::borrow::Cow;
-use std::marker::PhantomData;
 use thiserror::Error;
 
+use std::borrow::Cow;
+use std::collections::HashSet;
+use std::marker::PhantomData;
+
 #[derive(Clone)]
 pub enum Pattern<L: Language> {
   MetaVar {
@@ -125,6 +127,42 @@ impl<L: Language> Pattern<L> {
       _ => false,
     }
   }
+
+  /// Get all defined variables in the pattern.
+  /// Used for validating rules and report undefined variables.
+  pub fn defined_vars(&self) -> HashSet<&str> {
+    let mut vars = HashSet::new();
+    collect_vars(self, &mut vars);
+    vars
+  }
+}
+
+fn meta_var_name(meta_var: &MetaVariable) -> Option<&str> {
+  use MetaVariable as MV;
+  match meta_var {
+    MV::Capture(name, _) => Some(name),
+    MV::MultiCapture(name) => Some(name),
+    MV::Dropped(_) => None,
+    MV::Multiple => None,
+  }
+}
+
+fn collect_vars<'p, L: Language>(p: &'p Pattern<L>, vars: &mut HashSet<&'p str>) {
+  match p {
+    Pattern::MetaVar { meta_var, .. } => {
+      if let Some(name) = meta_var_name(meta_var) {
+        vars.insert(name);
+      }
+    }
+    Pattern::Terminal { .. } => {
+      // collect nothing for terminal nodes!
+    }
+    Pattern::Internal { children, .. } => {
+      for c in children {
+        collect_vars(c, vars);
+      }
+    }
+  }
 }
 
 impl<L: Language> Pattern<L> {
@@ -447,4 +485,33 @@ mod test {
       "[var, [Capture(\"A\", true), =, 1]]"
     );
   }
+
+  fn defined_vars(s: &str) -> Vec<String> {
+    let pattern = Pattern::str(s, Tsx);
+    let mut vars: Vec<_> = pattern
+      .defined_vars()
+      .into_iter()
+      .map(String::from)
+      .collect();
+    vars.sort();
+    vars
+  }
+
+  #[test]
+  fn test_extract_meta_var_from_pattern() {
+    let vars = defined_vars("var $A = 1");
+    assert_eq!(vars, ["A"]);
+  }
+
+  #[test]
+  fn test_extract_complex_meta_var() {
+    let vars = defined_vars("function $FUNC($$$ARGS): $RET { $$$BODY }");
+    assert_eq!(vars, ["ARGS", "BODY", "FUNC", "RET"]);
+  }
+
+  #[test]
+  fn test_extract_duplicate_meta_var() {
+    let vars = defined_vars("var $A = $A");
+    assert_eq!(vars, ["A"]);
+  }
 }
